@@ -32,18 +32,18 @@ sub add_data_series {
   return;
 }
 
-=item add_column_data_series(\@data, $series_name)
+=item add_bar_data_series(\@data, $series_name)
 
-Add a column data series to the graph.
+Add a bar data series to the graph.
 
 =cut
 
-sub add_column_data_series {
+sub add_bar_data_series {
   my $self = shift;
   my $data_ref = shift;
   my $series_name = shift;
 
-  $self->_add_data_series('column', $data_ref, $series_name);
+  $self->_add_data_series('bar', $data_ref, $series_name);
 
   return;
 }
@@ -101,6 +101,9 @@ sub draw {
 
   my $style = $self->{_style};
 
+  $self->_make_img
+    or return;
+
   my $img = $self->_get_image()
     or return;
 
@@ -116,7 +119,7 @@ sub draw {
 
   # Scale the graph box down to the widest graph that can cleanly hold the # of columns.
   return unless $self->_get_data_range();
-  $self->_remove_tics_from_chart_box(\@chart_box);
+  $self->_remove_tics_from_chart_box(\@chart_box, \%opts);
   my $column_count = $self->_get_column_count();
 
   my $width = $self->_get_number('width');
@@ -125,38 +128,49 @@ sub draw {
   my $graph_width = $chart_box[2] - $chart_box[0];
   my $graph_height = $chart_box[3] - $chart_box[1];
 
-  my $col_height = int(($graph_height - 1) / $column_count) -1;
-  $graph_height = $col_height * $column_count + 1;
+  my $col_height = ($graph_height - 1) / $column_count;
+  if ($col_height > 1) {
+    $graph_height = int($col_height) * $column_count + 1;
+  }
+  else {
+    $graph_height = $col_height * $column_count + 1;
+  }
 
   my $tic_count = $self->_get_x_tics();
   my $tic_distance = int(($graph_width -1) / ($tic_count - 1));
   $graph_width = $tic_distance * ($tic_count - 1);
 
-  my $bottom = $chart_box[1];
+  my $top = $chart_box[1];
   my $left   = $chart_box[0];
 
   $self->{'_style'}{'graph_width'} = $graph_width;
   $self->{'_style'}{'graph_height'} = $graph_height;
 
-  my @graph_box = ($left, $bottom, $left + $graph_width, $bottom + $graph_height);
+  my @graph_box = ($left, $top, $left + $graph_width, $top + $graph_height);
+
   $self->_set_graph_box(\@graph_box);
 
-  $img->box(
-            color   => $self->_get_color('outline.line'),
-            xmin    => $left,
-            xmax    => $left+$graph_width,
-            ymin    => $bottom,
-            ymax    => $bottom+$graph_height,
-            );
+  my @fill_box = @graph_box;
+
+  if ($self->_feature_enabled("graph_outline")) {
+    my @line = $self->_get_line("graph.outline")
+      or return;
+
+    $self->_box(
+		@line,
+		box => \@fill_box,
+		img => $img,
+	       );
+    ++$fill_box[0];
+    ++$fill_box[1];
+    --$fill_box[2];
+    --$fill_box[3];
+  }
 
   $img->box(
-            color   => $self->_get_color('bg'),
-            xmin    => $left + 1,
-            xmax    => $left+$graph_width - 1,
-            ymin    => $bottom + 1,
-            ymax    => $bottom+$graph_height-1 ,
-            filled  => 1,
-            );
+	    $self->_get_fill("graph.fill"),
+	    box => \@fill_box,
+	   );
 
   my $min_value = $self->_get_min_value();
   my $max_value = $self->_get_max_value();
@@ -164,7 +178,7 @@ sub draw {
 
   my $zero_position;
   if ($value_range) {
-    $zero_position =  $left + (-1*$min_value / $value_range) * ($graph_width-1);
+    $zero_position =  $left + $graph_width + (-1*$min_value / $value_range) * ($graph_width-1);
   }
 
   if ($min_value < 0) {
@@ -172,28 +186,33 @@ sub draw {
             color   => $self->_get_color('negative_bg'),
             xmin    => $left+1,
             xmax    => $zero_position,
-            ymin    => $bottom+1,
-            ymax    => $bottom+$graph_height - 1,
+            ymin    => $top+1,
+            ymax    => $top+$graph_height - 1,
             filled  => 1,
     );
     $img->line(
             x1 => $zero_position,
-            y1 => $bottom,
+            y1 => $top,
             x2 => $zero_position,
-            y2 => $bottom + $graph_height,
+            y2 => $top + $graph_height,
             color => $self->_get_color('outline.line'),
     );
   }
 
+  $self->_reset_series_counter();
+
   if ($self->_get_data_series()->{'bar'}) {
     $self->_draw_bars();
   }
+  if ($self->_get_data_series()->{'line'}) {
+    $self->_draw_lines();
+  }
 
   if ($self->_get_x_tics()) {
     $self->_draw_x_tics();
   }
-  if ($self->_get_labels()) {
-    $self->_draw_y_tics();
+  if ($self->_get_labels(\%opts)) {
+    $self->_draw_y_tics(\%opts);
   }
 
   return $self->_get_image();
@@ -207,12 +226,50 @@ sub _get_data_range {
   my $column_count = 0;
 
   my ($b_min, $b_max, $b_cols) = $self->_get_bar_range();
+  my ($l_min, $l_max, $l_cols) = $self->_get_line_range();
 
-  $min_value = $b_min;
-  $max_value = $b_max;
-  $column_count = $b_cols;
-#  $min_value = $self->_min(STARTING_MIN_VALUE, $c_min, $l_min);
-#  $max_value = $self->_max(0, $sc_max, $c_max, $l_max);
+  $min_value = $self->_min(STARTING_MIN_VALUE, $b_min, $l_min);
+  $max_value = $self->_max(0, $b_max, $l_max);
+  $column_count = $self->_max(0, $b_cols, $l_cols);
+
+  my $config_min = $self->_get_number('x_min');
+  my $config_max = $self->_get_number('x_max');
+
+  if (defined $config_max && $config_max < $max_value) {
+    $config_max = undef;
+  }
+  if (defined $config_min && $config_min > $min_value) {
+    $config_min = undef;
+  }
+
+  my $range_padding = $self->_get_number('range_padding');
+  if (defined $config_min) {
+    $min_value = $config_min;
+  }
+  else {
+    if ($min_value > 0) {
+      $min_value = 0;
+    }
+    if ($range_padding && $min_value < 0) {
+      my $difference = $min_value * $range_padding / 100;
+      if ($min_value < -1 && $difference > -1) {
+        $difference = -1;
+      }
+      $min_value += $difference;
+    }
+  }
+  if (defined $config_max) {
+    $max_value = $config_max;
+  }
+  else {
+    if ($range_padding && $max_value > 0) {
+      my $difference = $max_value * $range_padding / 100;
+      if ($max_value > 1 && $difference < 1) {
+        $difference = 1;
+      }
+      $max_value += $difference;
+    }
+  }
 
   if ($self->_get_number('automatic_axis')) {
     # In case this was set via a style, and not by the api method
@@ -260,6 +317,34 @@ sub _max {
   return $min;
 }
 
+sub _get_line_range {
+  my $self = shift;
+  my $series = $self->_get_data_series()->{'line'};
+  return (undef, undef, 0) unless $series;
+
+  my $max_value = 0;
+  my $min_value = STARTING_MIN_VALUE;
+  my $column_count = 0;
+
+  my @series = @{$series};
+  foreach my $series (@series) {
+    my @data = @{$series->{'data'}};
+
+    if (scalar @data > $column_count) {
+      $column_count = scalar @data;
+    }
+
+    foreach my $value (@data) {
+      if ($value > $max_value) { $max_value = $value; }
+      if ($value < $min_value) { $min_value = $value; }
+    }
+  }
+
+  return ($min_value, $max_value, $column_count);
+}
+
+
+
 sub _get_bar_range {
   my $self = shift;
 
@@ -330,14 +415,13 @@ sub _draw_lines {
 
   my $has_columns = (defined $self->_get_data_series()->{'column'} || $self->_get_data_series->{'stacked_column'}) ? 1 : 0;
 
-  my $col_width = int($graph_width / $column_count) -1;
+  my $col_height = int($graph_height / $column_count) -1;
 
   my $graph_box = $self->_get_graph_box();
   my $left = $graph_box->[0] + 1;
   my $bottom = $graph_box->[1];
 
-  my $zero_position =  $bottom + $graph_height - (-1*$min_value / $value_range) * ($graph_height - 1);
-
+  my $zero_position =  $left + $graph_width - (-1*$min_value / $value_range) * ($graph_width - 1);
 
   my $line_aa = $self->_get_number("lineaa");
   foreach my $series (@$line_series) {
@@ -346,37 +430,40 @@ sub _draw_lines {
 
     my $interval;
     if ($has_columns) {
-      $interval = $graph_width / ($data_size);
+      $interval = $graph_height / ($data_size);
     }
     else {
-      $interval = $graph_width / ($data_size - 1);
+      $interval = $graph_height / ($data_size - 1);
     }
     my $color = $self->_data_color($series_counter);
 
     # We need to add these last, otherwise the next line segment will overwrite half of the marker
     my @marker_positions;
     for (my $i = 0; $i < $data_size - 1; $i++) {
-      my $x1 = $left + $i * $interval;
-      my $x2 = $left + ($i + 1) * $interval;
+      my $y1 = $bottom + $i * $interval;
+      my $y2 = $bottom + ($i + 1) * $interval;
 
-      $x1 += $has_columns * $interval / 2;
-      $x2 += $has_columns * $interval / 2;
+      $y1 += $has_columns * $interval / 2;
+      $y2 += $has_columns * $interval / 2;
 
-      my $y1 = $bottom + ($value_range - $data[$i] + $min_value)/$value_range * $graph_height;
-      my $y2 = $bottom + ($value_range - $data[$i + 1] + $min_value)/$value_range * $graph_height;
+      my $x1 = $left + ($value_range - $data[$i] + $min_value)/$value_range * $graph_width;
+      my $x2 = $left + ($value_range - $data[$i + 1] + $min_value)/$value_range * $graph_width;
 
       push @marker_positions, [$x1, $y1];
       $img->line(x1 => $x1, y1 => $y1, x2 => $x2, y2 => $y2, aa => $line_aa, color => $color) || die $img->errstr;
     }
 
-    my $x2 = $left + ($data_size - 1) * $interval;
-    $x2 += $has_columns * $interval / 2;
 
-    my $y2 = $bottom + ($value_range - $data[$data_size - 1] + $min_value)/$value_range * $graph_height;
+    my $y2 = $bottom + ($data_size - 1) * $interval;
+    $y2 += $has_columns * $interval / 2;
 
-    push @marker_positions, [$x2, $y2];
-    foreach my $position (@marker_positions) {
-      $self->_draw_line_marker($position->[0], $position->[1], $series_counter);
+    my $x2 = $left + ($value_range - $data[$data_size - 1] + $min_value)/$value_range * $graph_width;
+
+    if ($self->_feature_enabled("linemarkers")) {
+      push @marker_positions, [$x2, $y2];
+      foreach my $position (@marker_positions) {
+	$self->_draw_line_marker($position->[0], $position->[1], $series_counter);
+      }
     }
     $series_counter++;
   }
@@ -385,76 +472,6 @@ sub _draw_lines {
   return;
 }
 
-sub _line_marker {
-  my ($self, $index) = @_;
-
-  my $markers = $self->{'_style'}{'line_markers'};
-  if (!$markers) {
-    return;
-  }
-  my $marker = $markers->[$index % @$markers];
-
-  return $marker;
-}
-
-sub _draw_line_marker {
-  my $self = shift;
-  my ($x1, $y1, $series_counter) = @_;
-
-  my $img = $self->_get_image();
-
-  my $style = $self->_line_marker($series_counter);
-  return unless $style;
-
-  my $type = $style->{'shape'};
-  my $radius = $style->{'radius'};
-
-  my $line_aa = $self->_get_number("lineaa");
-  my $fill_aa = $self->_get_number("fill.aa");
-  if ($type eq 'circle') {
-    my @fill = $self->_data_fill($series_counter, [$x1 - $radius, $y1 - $radius, $x1 + $radius, $y1 + $radius]);
-    $img->circle(x => $x1, y => $y1, r => $radius, aa => $fill_aa, filled => 1, @fill);
-  }
-  elsif ($type eq 'square') {
-    my @fill = $self->_data_fill($series_counter, [$x1 - $radius, $y1 - $radius, $x1 + $radius, $y1 + $radius]);
-    $img->box(xmin => $x1 - $radius, ymin => $y1 - $radius, xmax => $x1 + $radius, ymax => $y1 + $radius, @fill);
-  }
-  elsif ($type eq 'diamond') {
-    # The gradient really doesn't work for diamond
-    my $color = $self->_data_color($series_counter);
-    $img->polygon(
-        points => [
-                    [$x1 - $radius, $y1],
-                    [$x1, $y1 + $radius],
-                    [$x1 + $radius, $y1],
-                    [$x1, $y1 - $radius],
-                  ],
-        filled => 1, color => $color, aa => $fill_aa);
-  }
-  elsif ($type eq 'triangle') {
-    # The gradient really doesn't work for triangle
-    my $color = $self->_data_color($series_counter);
-    $img->polygon(
-        points => [
-                    [$x1 - $radius, $y1 + $radius],
-                    [$x1 + $radius, $y1 + $radius],
-                    [$x1, $y1 - $radius],
-                  ],
-        filled => 1, color => $color, aa => $fill_aa);
-
-  }
-  elsif ($type eq 'x') {
-    my $color = $self->_data_color($series_counter);
-    $img->line(x1 => $x1 - $radius, y1 => $y1 -$radius, x2 => $x1 + $radius, y2 => $y1+$radius, aa => $line_aa, color => $color) || die $img->errstr;
-    $img->line(x1 => $x1 + $radius, y1 => $y1 -$radius, x2 => $x1 - $radius, y2 => $y1+$radius, aa => $line_aa, color => $color) || die $img->errstr;
-  }
-  elsif ($type eq 'plus') {
-    my $color = $self->_data_color($series_counter);
-    $img->line(x1 => $x1, y1 => $y1 -$radius, x2 => $x1, y2 => $y1+$radius, aa => $line_aa, color => $color) || die $img->errstr;
-    $img->line(x1 => $x1 + $radius, y1 => $y1, x2 => $x1 - $radius, y2 => $y1, aa => $line_aa, color => $color) || die $img->errstr;
-  }
-}
-
 sub _draw_bars {
   my $self = shift;
   my $style = $self->{'_style'};
@@ -480,7 +497,7 @@ sub _draw_bars {
 
   my $zero_position =  int($left + (-1*$min_value / $value_range) * ($graph_width-1));
 
-  my $bar_height = int($graph_height / $column_count - 1);
+  my $bar_height = $graph_height / $column_count;
 
   my $outline_color;
   if ($style->{'features'}{'outline'}) {
@@ -500,9 +517,15 @@ sub _draw_bars {
     my $data_size = scalar @data;
     for (my $i = 0; $i < $data_size; $i++) {
 
-      my $y1 = int($bottom + $bar_height * (scalar @$col_series * $i + $series_pos)) + scalar @$col_series * $i + $series_pos + ($column_padding / 2);
+      my $part1 = $bar_height * (scalar @$col_series * $i);
+      my $part2 = ($series_pos) * $bar_height;
+      my $y1 = int($bottom + $part1 + $part2);
 
-      my $y2 = $y1 + $bar_height - $column_padding;
+      my $y2 = int($y1 + $bar_height - $column_padding)-1;
+      # Special case for when bar_height is less than 1.
+      if ($y2 < $y1) {
+        $y2 = $y1;
+      }
 
       my $x1 = int($left - ($min_value - $data[$i]) / $value_range * $graph_width);
 
@@ -557,10 +580,51 @@ sub _add_data_series {
 
 Shows vertical gridlines at the y-tics.
 
+Feature: vertical_gridlines
+
 =cut
 
 sub show_vertical_gridlines {
-    $_[0]->{'custom_style'}->{'vertical_gridlines'} = 1;
+    $_[0]->{'custom_style'}{features}{'vertical_gridlines'} = 1;
+}
+
+=item set_vertical_gridline_style(color => ..., style => ...)
+
+Set the color and style of the lines drawn for gridlines.
+
+Style equivalent: vgrid
+
+=cut
+
+sub set_vertical_gridline_style {
+  my ($self, %opts) = @_;
+
+  $self->{custom_style}{vgrid} ||= {};
+  @{$self->{custom_style}{vgrid}}{keys %opts} = values %opts;
+
+  return 1;
+}
+
+=item show_line_markers()
+
+=item show_line_markers($value)
+
+Feature: linemarkers.
+
+If $value is missing or true, draw markers on a line data series.
+
+Note: line markers are drawn by default.
+
+=cut
+
+sub show_line_markers {
+  my ($self, $value) = @_;
+
+  @_ > 1 or $value = 1;
+
+  $self->{custom_style}{features}{linemarkers} = $value;
+
+  return 1;
 }
 
 =item use_automatic_axis()
@@ -594,11 +658,10 @@ sub _get_x_tics {
 }
 
 sub _remove_tics_from_chart_box {
-  my $self = shift;
-  my $chart_box = shift;
+  my ($self, $chart_box, $opts) = @_;
 
   # XXX - bad default
-  my $tic_width = $self->_get_y_tic_width() || 10;
+  my $tic_width = $self->_get_y_tic_width($opts) || 10;
   my @y_tic_box = ($chart_box->[0], $chart_box->[1], $chart_box->[0] + $tic_width, $chart_box->[3]);
 
   # XXX - bad default
@@ -612,12 +675,31 @@ sub _remove_tics_from_chart_box {
   my @y_tic_tops = ($chart_box->[0], $chart_box->[1], $chart_box->[2], $chart_box->[1] + int($tic_height / 2));
   $self->_remove_box($chart_box, \@y_tic_tops);
 
+    if (my @box = $self->_text_bbox($self->_get_max_value(), 'legend')) {
+      my @remove_box = ($chart_box->[2] - int($box[2] / 2) - 1,
+                        $chart_box->[1],
+                        $chart_box->[2],
+                        $chart_box->[3]
+                        );
+
+      $self->_remove_box($chart_box, \@remove_box);
+    }
+
+
 }
 
 sub _get_y_tic_width {
-  my $self = shift;
+  my ($self, $opts) = @_;
+
+  my $labels = $self->_get_labels($opts);
+
+  if (!$labels) {
+    return;
+  }
+
+  my %text_info = $self->_text_style('legend')
+    or return;
 
-  my $labels = $self->_get_labels();
   my $max_width = 0;
   foreach my $label (@$labels) {
     my @box = $self->_text_bbox($label, 'legend');
@@ -662,13 +744,13 @@ sub _get_x_tic_height {
 }
 
 sub _draw_y_tics {
-  my $self = shift;
+  my ($self, $opts) = @_;
 
   my $img = $self->_get_image();
   my $graph_box = $self->_get_graph_box();
   my $image_box = $self->_get_image_box();
 
-  my $labels = $self->_get_labels();
+  my $labels = $self->_get_labels($opts);
 
   my $tic_count = (scalar @$labels) - 1;
 
@@ -731,7 +813,8 @@ sub _draw_x_tics {
   my %text_info = $self->_text_style('legend')
     or return;
 
-  my $show_gridlines = $self->_get_number('vertical_gridlines');
+  my $show_gridlines = $self->{_style}{features}{'vertical_gridlines'};
+  my @grid_line = $self->_get_line("vgrid");
   for my $count (0 .. $tic_count-1) {
     my $x1 = $graph_box->[0] + ($tic_distance * $count);
 
@@ -754,14 +837,11 @@ sub _draw_x_tics {
                  text => $value
                 );
 
-    if ($show_gridlines) {
-      # XXX - line styles!
-      for (my $i = $graph_box->[1]; $i < $graph_box->[3]; $i += 6) {
-        my $y1 = $i;
-        my $y2 = $i + 2;
-        if ($y2 > $graph_box->[3]) { $y2 = $graph_box->[3]; }
-        $img->line(x1 => $x1, x2 => $x1, y1 => $y1, y2 => $y2, aa => 1, color => '000000');
-      }
+    if ($show_gridlines && $x1 != $graph_box->[0] && $x1 != $graph_box->[2]) {
+      $self->_line(x1 => $x1, x2 => $x1,
+		   y1 => $graph_box->[1], y2 => $graph_box->[3],
+		   img => $img,
+		   @grid_line);
     }
   }
 }
@@ -798,7 +878,27 @@ sub _get_min_value      { return $_[0]->{'min_value'} }
 sub _get_max_value      { return $_[0]->{'max_value'} }
 sub _get_image_box      { return $_[0]->{'image_box'} }
 sub _get_graph_box      { return $_[0]->{'graph_box'} }
+sub _reset_series_counter { $_[0]->{series_counter} = 0 }
 sub _get_series_counter { return $_[0]->{'series_counter'} }
 
+sub _style_defs {
+  my ($self) = @_;
+
+  my %work = %{$self->SUPER::_style_defs()};
+  push @{$work{features}}, qw/graph_outline graph_fill linemarkers/;
+  $work{vgrid} =
+    {
+     color => "lookup(fg)",
+     style => "solid",
+    };
+
+  return \%work;
+}
+
+sub _composite {
+  my ($self) = @_;
+  return ( $self->SUPER::_composite(), "graph", "vgrid" );
+}
+
 1;
 
